/**********************************************************************
                               Macro LORG
AUTHOR:
     Zhiwu Zhang (zz19@cornell.edu), May 2004.

PURPOSE:
     Create genetic (cov)variance  for SAS PROC MIXED.

REQUIRES:
     Requires Base SAS and SAS/IML software, Version 8 or later.

INPUT:
	DataRelation -	SAS dataset of K or A matrix (up triangle).
	DataRandom -	SAS dataset of random variables used in RANDOM statement
	DATALORG -	SAS dataset contains variance and covariance matrix of random viable 
				across traits if these parameters are known. Otherwise, the elements 
				could be any numbers except the option for user to define these parameters.
	LORGType -	Parameter for the status of variance and covariance parameters: "G" 
				for known, "A" for unknown and assign parameters automatically, "U" 
				for unknown and assign parameters by user defined in SAS dataset DATALORG.
OUTPUT:
	DataG -			SAS dataset The columns depends on LORGType. The columns are ROW, COL 
				and VALUE if LORGType="G". Otherwise, the columns are ROW, COL, VALUE and PARM.
/*********************************************************************/

%macro LORG(DataRelation=,DataRandom=,dataLorG=,LORGType=,DataG=);
proc iml;
start main;

*Step 1: Convert sas dataset to IML array;

*Import K or A matrix;
use &DataRelation;
read all into AAA;
nid = nrow(AAA);

*Import Gnull;
use &dataLorG;
read all into g;
sizeofGnull = nrow(g);

*Import random variable;
use &DataRandom;
read all into list;

*Step 2: Configration;
nobs = nrow(list);
r = ncol(list);
N = j(r,1,0);*vector for number of ind for each random variable;

*Get a coresponding to random effects to be used in proc mixed;
numInd=0;
kk = j(nobs,r,0);*array for random variable;
do mr=1 to r;
	k=unique(list[1:nobs,mr]);
	N[mr] = nrow(k`);
	if(k[1]=.|k[1]=0)then
		do;
			k=k[2:N[mr]];
			N[mr]=N[mr]-1;
		end;
	kk[1:N[mr],mr]=k[1:N[mr]];
	numInd=numInd+N[mr];
end;

*get number of trait;
t=sizeofGnull/r;
if (t<1) then 
	do;
		print 'G does not match random variables';
	t=1;
end;

*Step 3: Build parm array;
parm=j(t*r,t*r,0);
do mr=1 to r;*loop for row of random variable mr: row wise;
	do mc=1 to r;*loop for col of random variable mc: col wise;
		do i=1 to t;*loop on trait i: row wise;
			do j=1 to t;*loop on trait i: col wise;
				Iindex=(mr-1)*t+i;
				Jindex=(mc-1)*t+j;
				if (Iindex<=Jindex) then
					fp=Jindex*(Jindex-1)/2+Iindex;
				else
					fp=Iindex*(Iindex-1)/2+Jindex;
				parm[Iindex,Jindex]=fp;
				parm[Jindex,Iindex]=fp;
			end;
		end;
	end;
end;

*Step 4: Build LorGdata;
if(&LorGType='G'|&LorGType='g') then
	L=j(numInd*numInd*t*t,3,0);
else
	L=j(numInd*numInd*t*t,4,0);
fq=0;
rcum=0;
do mr=1 to r;*loop for row of random variable mr: row wise;
	ccum=0;
	do mc=1 to r;*loop for col of random variable mc: col wise;
		frKept=rcum;
		kkk=j(N[mr]*N[mc],1,0);
		do halfi=1 to N[mr];
			do halfj=1 to N[mc];
				if(kk[halfi,mr]>kk[halfj,mc]) then
					kkk[N[mc]*(halfi-1)+halfj]=kk[halfi,mr]*(kk[halfi,mr]-1)/2+kk[halfj,mc];
				else
					kkk[N[mc]*(halfi-1)+halfj]=kk[halfj,mc]*(kk[halfj,mc]-1)/2+kk[halfi,mr];
			end;
		end;
		AA=AAA[kkk];
		do i=1 to t;*loop on trait i: row wise;
			fcKept=ccum;
			do j=1 to t;*loop on trait i: col wise;
				fr=frKept;
				*set parm;
				Iindex=(mr-1)*t+i;
				Jindex=(mc-1)*t+j;
				if(&LorGType='A'|&LorGType='a') then
					fp=parm[Iindex,Jindex];
				else if (&LorGType='G'|&LorGType='g') then
						fp=0;*Gdata do not requir parm;
					else
						fp=g[Iindex,Jindex];

				do s=1 to N[mr];*loop on ind s: row wise;
					fr=fr+1;
					fc=fcKept;
					do d=1 to N[mc];*loop on ind d: col wise;
						positionSD=N[mc]*(s-1)+d;

						fc=fc+1;
						if(&LorGType='G'|&LorGType='g') then
							fv=g[Iindex,Jindex]*AA[positionSD];
						else
							fv=AA[positionSD];
						if(fv>0)then;
							do;
								fq=fq+1;
								L[fq,1]=fr;
								L[fq,2]=fc;
								L[fq,3]=fv;
								if(^(&LorGType='G'|&LorGType='g')) then
									L[fq,4]=fp;
							end;
					end;*end of loop on ind d;
				end;*end of loop on ind s;
				fcKept=fcKept+N[mc];
			end;*end of loop on trait j;
			frKept=frKept+N[mr];
		end;*end of loop on trait i;
		ccum=fckept;
	end;*end of loop for row of random variable mc ;
	rcum=frkept;
end;*end of loop for col of random variable mr;

*Step 5: Output;
if(&LorGType='G'|&LorGType='g') then
	cnames={row col value};
else
	cnames={row col value parm };
if(&LorGType='G'|&LorGType='g') then
	L=L[1:fq,1:3];
else
	L=L[1:fq,1:4];
create &DataG from L[colname=cnames] ;
append from L;

finish;
run main;
quit;
%mend;
